#include "Markov_stride.h"

/*
  This code builds a framework that helps model a program's behaviour. It is programmed
  in such a way that is should seem to the user as being a black-box. Not a lot of functions
  are needed to be able to make it run.

  For any problem of installation or usage, please email me at: beyler@dpt-info.u-strasbg.fr

  Copyright (C) 2006 Beyler Jean Christophe

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

inline static void Markov_InlinePrefetch(void *address)
{
__asm__ __volatile__ (
	"lfetch [%0]\n"
	//"ld1.s r31=[%0]\n"
	::"r" (address)
               );
}

static void Markov_Wait(SMarkov* m, void *address)
{
	int stride;
	#ifdef __ADDRESS_BLOCK__
		#ifdef __POINTER_32_BIT__
			{
			int add = (int) address;
			add &= __ADDRESS_BLOCK__;
			address = (void *) add;
			}
		#else
			{
			long add = (long) address;
			add &= __ADDRESS_BLOCK__;
			address = (void*) add;
			}
		#endif
	#endif

	//Calculate stride
	stride = address - m->address;
	//Set base address
	m->address = address;
	
	//Check our accuracy
	#ifdef __OPT_VERIF__
		Markov_CheckPred(m,address);
	        m->cnt_calls++;
		if(stride==0)
			m->nbr_of_0s++;		
	#endif

	#ifdef __DEBUG__
		printf("Module markov Wait\n");
	#endif

	//if tmp is NULL, we see if there is not a root node with the correct info
	//search for a node labeled info in the binary tree (Markov_SearchNode returns the associated prediction node)
	m->cur[0] = Markov_SearchNode(stride,m->root);
	if(m->cur[0])
		m->fct = &Markov_Predict;

	//increment the number of consecutive mistakes
  	m->nbr_err++;
	//if this number exceeds a threshold, we flush the graph and restart
	if(m->nbr_err>m->nbr_err_max)
		{
		#ifdef __NORMAL_VERB__
			printf("A flush\n");
		#endif
		Markov_Reset(m);
		return;
		}
}

//predict is the function for the prediction phase
inline static void Markov_Inline(SMarkov* m,void* address)
{
	SNode* tmp ,*tmp2;
	SNodeList* ltmp;

	int stride;
	int max_vis;
	int itmp;

	if(m->fct != &Markov_Predict)
		{
		m->fct(m,address);
		return;
		}

	#ifdef __ADDRESS_BLOCK__
		#ifdef __POINTER_32_BIT__
			{
			int add = (int) address;
			add &= __ADDRESS_BLOCK__;
			address = (void *) add;
			}
		#else
			{
			long add = (long) address;
			add &= __ADDRESS_BLOCK__;
			address = (void*) add;
			}
		#endif
	#endif

	//Calculate stride
	stride = address - m->address;
	//Set base address
	m->address = address;
	
	//Check our accuracy
	#ifdef __OPT_VERIF__
		Markov_CheckPred(m,address);
	        m->cnt_calls++;
		if(stride==0)
			m->nbr_of_0s++;		
	#endif

	#ifdef __DEBUG__
		printf("Module markov Predict\n");
	#endif

	//Our position in the graph of prediction is always stored in m->cur[0]
	tmp = m->cur[0];

	{	//Otherwise tmp is not NULL, let us see if we have a son with the stride stride

		/* Each node has three special pointers:
		*	probableNext: that is the most often followed son
		*	last and last2: which are the two last nodes followed
		
		*  And three integers:
		*	vis_probableNext, vis_last and vis_last2: These represent the number of times
		*	we followed the associated edges.
		*	If ever, one of these nodes disappears, we update the associated arc
		*/

		//get tmp->probableNext and vis_probableNext
		tmp2 = tmp->probableNext;
		
		//If it is the most often followed node
		if(tmp2->stride==stride)
			{
			//We update m->cur[0], it is now worth the most followed son of the current node
			m->cur[0] = tmp2;

			//we reset the number of mistakes
			m->nbr_err = 0;
			
			//Update visit
			tmp->vis_probableNext++;			
			
			#ifdef __VERSION_CHECK__
				//Final check before we forget this node	
				if(tmp->node_ver<tmp2->node_ver)
					{
					Markov_CalcStridePred(m->prefDist,tmp);
					tmp->node_ver = tmp2->node_ver;

					#ifdef __OPT_VERIF__
						m->cnt_vc_nodeupdate++;
					#endif
					}	
			#endif

			#ifdef __OPT_VERIF__
				//If we want to check the accuracy of the model
				Markov_CommitPred(m,address+tmp2->predStride);
			#endif

			#ifdef __PREFETCH__
				//we prefetch the info
				Markov_InlinePrefetch(address+tmp2->predStride);
			#endif
			return;
			}

		//if we are here then it is not the most followed edge,
		//perhaps it is the last node followed?
		max_vis = tmp->vis_probableNext;
		tmp2 = tmp->last;
		
		if(tmp2->stride==stride)
			{
			//if so, the current node becomes this one
			m->cur[0] = tmp2;
			
			//Update visit
			tmp->vis_last++;
			
			if(tmp->vis_last>max_vis)
					{
					tmp->last = tmp->probableNext;
					//tmp2 is equal to tmp->last
					tmp->probableNext = tmp2;

					//let us switch the vis_probableNext and vis_last variable
					tmp->vis_probableNext = tmp->vis_last;
					tmp->vis_last = max_vis;

					#ifdef __VERSION_CHECK__			
						//Update node version
						tmp->node_ver++;
						#ifdef __OPT_VERIF__
							m->cnt_node_ver_upd++;
						#endif
					#endif

					//recalculate the prediction
					Markov_CalcStridePred(m->prefDist,tmp);
					}
			#ifdef __OPT_VERIF__
				//If we want to check the accuracy of the model
				Markov_CommitPred(m,address+tmp2->predStride);
			#endif

			//we increment the number of mistakes
			m->nbr_err++;
			//if nbr_err is over the threshold
			if(m->nbr_err>m->nbr_err_max)
				{
				#ifdef __NORMAL_VERB__
					printf("Flushing\n");
				#endif
				Markov_Reset(m);
				return;
				}
			//we predict
			#ifdef __PREFETCH__
				Markov_InlinePrefetch(tmp2->predStride+address);
			#endif
			return;
			}

                
		//if we still have not found the son than we try with this one
		tmp2 = tmp->last2;

		//Check if there is an equality
		if(tmp2->stride==stride)
			{
			//we update m->cur[0]
			m->cur[0] = tmp2;

			//we increment the number of visits made to this node
			tmp->vis_last2++;

			//if ever vis_last2 is bigger than the threshold given by the most probable node
			//we switch
			if(tmp2->vis_last2>max_vis)
				{
				#ifdef __VERSION_CHECK__
					//Update node version
					tmp->node_ver++;
					#ifdef __OPT_VERIF__
						m->cnt_node_ver_upd++;
					#endif
				#endif

				//tmp2 is worth tmp->last2
				tmp->last2 = tmp->last;
				tmp->last = tmp->probableNext;
				tmp->probableNext = tmp2;

				//Let us switch the vis_probableNext, vis_last and vis_last2 variables
				tmp->vis_probableNext = tmp->vis_last2;
				tmp->vis_last2 = tmp->vis_last;
				tmp->vis_last = max_vis;

				//recalculate the full stride
				Markov_CalcStridePred(m->prefDist,tmp);
				}
							
			//increment the number of consecutive mistakes
			m->nbr_err++;
			//if nbr_err is over the threshold
			if(m->nbr_err>m->nbr_err_max)
				{
				#ifdef __NORMAL_VERB__
					printf("A Flush\n");
				#endif
				Markov_Reset(m);
				return;
				}

			#ifdef __OPT_VERIF__
				//If we want to check the accuracy of the model
				Markov_CommitPred(m,tmp2->predStride+address);
			#endif

			//we predict
			#ifdef __PREFETCH__
				Markov_InlinePrefetch(tmp2->predStride+address);
			#endif
			return;
			}

		//We check the other sons
		ltmp = tmp->next;

		while(ltmp!=NULL)
			{
			#ifdef __OPT_VERIF__
				//if we want more information
				m->cnt_jumps++;
			#endif

			//if the stride of this node is higher or equal we break
			if(ltmp->val->stride>=stride)
				break;
			ltmp = ltmp->next;
			}

		//if we are at the end or if it is not the right stride
		if((!ltmp)||(ltmp->val->stride!=stride))
			{
			
			//we increment the number of mistakes
			m->nbr_err++;
			if(m->nbr_err>m->nbr_err_max)
				{
				#ifdef __NORMAL_VERB__
					printf("Flushing\n");
				#endif
				Markov_Reset(m);
				return;
				}

			//we try to find a root node with the same label
   			m->cur[0] = Markov_SearchNode(stride,m->root);
			if(!(m->cur[0]))
				{
				m->fct = &Markov_Wait;
				}

			//if we do find one, we will use tmp->last to link us to it
			if(m->cur[0])
				{
				//before losing vis_last2, we try to see if last2 is a son of this node

				ltmp = tmp->next;
				while(ltmp!=NULL)
					{
					if(ltmp->val==tmp->last2)
						{
						//if we have found it
						ltmp->visit = tmp->vis_last2;
						break;
						}
					//Next edge
					ltmp = ltmp->next;
					}

				//we change the pointers
				tmp->last2 = tmp->last;
				tmp->last = m->cur[0];
				//we change vis_last2
				tmp->vis_last2 = tmp->vis_last;
				tmp->vis_last++;
				
				/* We do not change tmp->vis_last (we just increment it so that it has a minimum of 1)
				  because if we reset it to 0 we might
				  be much lower than tmp->vis_probableNext and so we would jeopardize this node becoming
				  the most followed one...

				    Leaving it, gives us the chance to be in the same ball park than the vis_last2 and vis_probableNext
				*/
				}
			}
		else //Otherwise we have a son with the right stride
			{
			tmp2 = ltmp->val;
			//we update the current pointer
			m->cur[0] = tmp2;

			//we update the vis_last2 and vis_last variables
			itmp = tmp->vis_last2;
			tmp->vis_last2 = tmp->vis_last;
			tmp->vis_last =ltmp->visit;

			//before losing itmp, we check to see if tmp->last2 is a son of this node
			ltmp = tmp->next;
			while(ltmp!=NULL)
				{
				if(ltmp->val==tmp->last2)
					{
					ltmp->visit = itmp;
					break;
					}
				ltmp = ltmp->next;
				}

			//update the pointers
			tmp->last2 = tmp->last;
			tmp->last = tmp2;

			#ifdef __OPT_VERIF__
				//If we want to check the accuracy of the model
				Markov_CommitPred(m,tmp2->predStride+address);
			#endif

			//prefetch the base address + the full stride
			#ifdef __PREFETCH__
				Markov_InlinePrefetch(tmp2->predStride+address);
			#endif
			}
	}
}

